<?php

namespace Drupal\monarch_virtual_folder\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Url;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Response;

/**
 * Returns responses for Monarch Virtual Folder routes.
 */
final class FolderApiController extends ControllerBase {

  /**
   * The database service.
   */
  private readonly Connection $database;

  /**
   * The file_url_generator service.
   */
  private readonly FileUrlGeneratorInterface $fileUrlGenerator;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    $instance = new self();

    $instance->database = $container->get('database');
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->currentUser = $container->get('current_user');
    $instance->moduleHandler = $container->get('module_handler');
    $instance->fileUrlGenerator = $container->get('file_url_generator');
    $instance->configFactory = $container->get('config.factory');

    return $instance;
  }

  /**
   * Prune folders with no files.
   */
  private function prune(array &$data, bool $prune_all = FALSE) {
    foreach ($data as $id => &$subdata) {
      if ($subdata['type'] === 'folder') {
        if (empty($subdata['children'])) {
          $subdata['parent'] = NULL;
          unset($data[$id]);
        }
        elseif ($prune_all || !$subdata['status']) {
          $subdata['parent'] = NULL;
          unset($data[$id]);
          $this->prune($subdata['children'], TRUE);
        }
      }
      elseif ($prune_all || is_null($subdata['parent'] ?? NULL) || $subdata['parent'] === 'folder:0') {
        $subdata['parent'] = NULL;
        unset($data[$id]);
      }
    }
  }

  /**
   * Generate a query for files this user can access.
   */
  private function fileQuery() {
    $media_query = $this->database->select('media_field_data', 'm')->condition('m.status', 1);

    $media_query->addExpression('\'media\'', 'type');
    $media_query->addField('m', 'mid', 'id');
    $media_query->addField('m', 'vid', 'revision');
    $media_query->addField('m', 'bundle');
    $media_query->addField('m', 'name');
    $media_query->addField('m', 'langcode');
    $media_query->addField('m', 'folder', 'parent');
    $media_query->addExpression('m.thumbnail__target_id', 'thumbnail');

    $node_query = $this->database->select('node_field_data', 'n')->condition('n.status', 1);

    $node_query->addExpression('\'node\'', 'type');
    $node_query->addField('n', 'nid', 'id');
    $node_query->addField('n', 'vid', 'revision');
    $node_query->addField('n', 'type', 'bundle');
    $node_query->addField('n', 'title', 'name');
    $node_query->addField('n', 'langcode');
    $node_query->addField('n', 'folder', 'parent');
    $node_query->addExpression('NULL', 'thumbnail');

    $mids = [];
    $nids = [];

    foreach ([
      'virtual_folder_media_ids' => &$mids,
      'virtual_folder_node_ids' => &$nids,
    ] as $hook => &$ids) {
      foreach ($this->moduleHandler()->invokeAll($hook) as $list) {
        if (!is_array($list)) {
          $list = [$list];
        }

        foreach ($list as $list_item) {
          if (is_numeric($list_item)) {
            $ids[$list_item] = $list_item;
          }
        }
      }
    }

    if (!empty($mids)) {
      $media_query->condition('mid', array_values($mids), 'IN');
    }

    if (!empty($nids)) {
      $node_query->condition('nid', array_values($nids), 'IN');
    }

    $query_wrapper = $this->database->select($media_query->union($node_query), 'f')->fields('f', [
      'id',
      'name',
      'langcode',
      'parent',
      'type',
      'thumbnail',
      'revision',
      'bundle',
    ]);

    $query_wrapper->addExpression('LOWER(name)', 'sort_name');
    $query_wrapper->orderBy('sort_name', 'ASC');

    return $query_wrapper;
  }

  /**
   * Builds the response.
   */
  public function data(): Response {
    if (!$this->currentUser || !$this->currentUser->isAuthenticated()) {
      return new JsonResponse(NULL, 404);
    }

    $start = hrtime(TRUE);

    $ret = [];

    $ret['folder:0'] = [
      'id' => 'folder:0',
      'type' => 'folder',
      'name' => 'Top',
      'langcode' => 'en',
      'parent' => 'folder:0',
      'status' => 1,
      'children' => [],
    ];

    $terms = $this->database->select('taxonomy_term_field_data', 't')
      ->fields('t', ['tid', 'name', 'langcode', 'status'])
      ->condition('vid', 'folders')
      ->orderBy('weight', 'ASC')
      ->execute();

    foreach ($terms as $term) {
      $ret['folder:' . $term->tid]['id'] = 'folder:' . $term->tid;
      $ret['folder:' . $term->tid]['name'] = $term->name;
      $ret['folder:' . $term->tid]['langcode'] = $term->langcode;
      $ret['folder:' . $term->tid]['parent'] = NULL;
      $ret['folder:' . $term->tid]['type'] = 'folder';
      $ret['folder:' . $term->tid]['status'] = intval($term->status);
      $ret['folder:' . $term->tid]['children'] = [];
    }

    $res = $this->database->select('taxonomy_term__parent', 't')
      ->fields('t', ['entity_id', 'parent_target_id'])
      ->condition('bundle', 'folders')
      ->execute();

    foreach ($res as $row) {
      $ret['folder:' . $row->entity_id]['parent'] = 'folder:' . $row->parent_target_id;
    }

    foreach ($ret as $tid => &$term_data) {
      $parent = $term_data['parent'];

      if ($tid !== $parent) {
        $ret[$parent]['children'][$tid] = &$term_data;
      }
    }

    foreach ($this->fileQuery()->execute() as $row) {
      $entity_type_id = $row->type;
      $id = $entity_type_id . ':' . $row->id;

      $ret[$id] = (array) $row;
      $ret[$id]['id'] = $id;
      $ret[$id]['type'] = $entity_type_id . ':' . $row->bundle;
      $ret[$id]['parent'] = NULL;

      if ($entity_type_id === 'node') {
        try {
          $url = Url::fromRoute('entity.node.canonical', [
            'node' => $row->id,
          ], [
            'absolute' => TRUE,
          ]);

          $ret[$id]['path'] = $url->toString();
        }
        catch (\Throwable $err) {
          // Ignore.
        }
      }
      elseif ($entity_type_id === 'media') {
        try {
          $url = Url::fromRoute('entity.media.canonical', [
            'media' => $row->id,
          ], [
            'absolute' => TRUE,
          ]);

          $ret[$id]['path'] = $url->toString();
        }
        catch (\Throwable $err) {
          // Ignore.
        }
      }

      unset($ret[$id]['revision']);
      unset($ret[$id]['bundle']);

      if (!empty($row->parent)) {
        $parent = 'folder:' . $row->parent;

        if (isset($ret[$parent])) {
          $ret[$id]['parent'] = $parent;
          $ret[$parent]['children'][$id] = &$ret[$id];
        }
        elseif ($id !== 'folder:0') {
          // If orphaned, don't set parent or children. Remove from list.
          unset($ret[$id]);
        }
      }
    }

    $this->prune($ret);

    $ret['folder:0']['parent'] = NULL;

    $thumbnails = [];

    foreach ($ret as $id => &$entry) {
      if ($entry['thumbnail'] ?? NULL) {
        $thumbnails[$entry['thumbnail']] = $entry['thumbnail'];
      }

      if (isset($entry['children'])) {
        $entry['children'] = array_keys($entry['children']);
      }
    }

    $uris = [];

    $thumbnail_style_id = $this->configFactory->get('monarch_virtual_folder.settings')->get('thumbnail_style_id') ?: 'folder_thumbnail';

    $image_style_storage = $this->entityTypeManager->getStorage('image_style');

    /** @var \Drupal\image\ImageStyleInterface|null $thumbnail_style */
    $thumbnail_style = $image_style_storage->load($thumbnail_style_id) ??
      $image_style_storage->load('folder_thumbnail') ??
      $image_style_storage->load('thumbnail');

    $generic_uri = ($this->configFactory->get('media.settings')->get('icon_base_uri') ?: 'public://media-icons/generic') . '/generic.png';
    $generic_url = $thumbnail_style ? $thumbnail_style->buildUrl($generic_uri) : $this->fileUrlGenerator->generateAbsoluteString($generic_uri);

    if (!empty($thumbnails)) {
      $uris = $this->database->select('file_managed', 'f')
        ->fields('f', ['fid', 'uri'])
        ->condition('fid', array_values($thumbnails), 'IN')
        ->execute()
        ->fetchAllKeyed();

      if ($thumbnail_style) {
        foreach ($uris as $fid => $uri) {
          $thumbnails[$fid] = $thumbnail_style->buildUrl($uri);
        }
      }
      else {
        foreach ($uris as $fid => $uri) {
          $thumbnails[$fid] = $this->fileUrlGenerator->generateAbsoluteString($uri);
        }
      }
    }

    foreach ($ret as $id => &$entry) {
      if (array_key_exists('thumbnail', $entry)) {
        if ($entry['thumbnail'] ?? NULL) {
          $entry['thumbnail'] = $thumbnails[$entry['thumbnail']] ?? NULL;
        }

        $entry['thumbnail'] = $entry['thumbnail'] ?? $generic_url;
      }
    }

    $this->moduleHandler->alter('virtual_folder_data', $ret);

    $end = hrtime(TRUE);

    return new JsonResponse((object) ($ret), 200, [
      'X-Elapsed' => ($end - $start) / 1000000,
    ]);
  }

}
